/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@cs.tu-berlin.de)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

/*  This is a literal translation of the Kernel Ridge Regression
    Matlab code of Dr Gavin C. Cawley, which can be found at

    http://theoval.sys.uea.ac.uk/~gcc/matlab/hkrr_demo/hkrr_demo.m

    I only translated the KRR part, not the HKRR. The translation is
    easily done using the array defined in "array.h". This code contains
    the original Matlab code as comments.

    Compilation is easiest on Linux with lapack (and perhaps gnuplot
    to see the output) installed. It also compiles under MSVC 6.0 (when
    you have lapack installed).

    Have fun! */

#include <math.h>
#include <fstream>
#include <array.h>

// fwd declarations
void rbf(arr& K,double lambda, const arr& Xbatch, const arr& x);
void krr(arr& alpha,double& b,double& sigma,
         double lambda, double gamma, const arr& x, const arr& t);

int main(int argn,char**argv){
  arr data;
  data.setText(
"      [  2.4    0.0 ;  2.6   -1.3 ;  3.2   -2.7 ;  3.6    0.0 ; \
          4.0   -2.7 ;  6.2   -2.7 ;  6.6   -2.7 ;  6.8   -1.3 ;\
          7.8   -2.7 ;  8.2   -2.7 ;  8.8   -1.3 ;  8.8   -2.7 ;\
          9.6   -2.7 ; 10.0   -2.7 ; 10.2   -5.4 ; 10.6   -2.7 ;\
         11.0   -5.4 ; 11.4    0.0 ; 13.2   -2.7 ; 13.6   -2.7 ;\
         13.8    0.0 ; 14.6  -13.3 ; 14.6   -5.4 ; 14.6   -5.4 ;\
         14.6   -9.3 ; 14.6  -16.0 ; 14.6  -22.8 ; 14.8   -2.7 ;\
         15.4  -22.8 ; 15.4  -32.1 ; 15.4  -53.5 ; 15.4  -54.9 ;\
         15.6  -40.2 ; 15.6  -21.5 ; 15.8  -21.5 ; 15.8  -50.8 ;\
         16.0  -42.9 ; 16.0  -26.8 ; 16.2  -21.5 ; 16.2  -50.8 ;\
         16.2  -61.7 ; 16.4   -5.4 ; 16.4  -80.4 ; 16.6  -59.0 ;\
         16.8  -71.0 ; 16.8  -91.1 ; 16.8  -77.7 ; 17.6  -37.5 ;\
         17.6  -85.6 ; 17.6 -123.1 ; 17.6 -101.9 ; 17.8  -99.1 ;\
         17.8 -104.4 ; 18.6 -112.5 ; 18.6  -50.8 ; 19.2 -123.1 ;\
         19.4  -85.6 ; 19.4  -72.3 ; 19.6 -127.2 ; 20.2 -123.1 ;\
         20.4 -117.9 ; 21.2 -134.0 ; 21.4 -101.9 ; 21.8 -108.4 ;\
         22.0 -123.1 ; 23.2 -123.1 ; 23.4 -128.5 ; 24.0 -112.5 ;\
         24.2  -95.1 ; 24.2  -81.8 ; 24.6  -53.5 ; 25.0  -64.4 ;\
         25.0  -57.6 ; 25.4  -72.3 ; 25.4  -44.3 ; 25.6  -26.8 ;\
         26.0   -5.4 ; 26.2 -107.1 ; 26.2  -21.5 ; 26.4  -65.6 ;\
         27.0  -16.0 ; 27.2  -45.6 ; 27.2  -24.2 ; 27.2    9.5 ;\
         27.6    4.0 ; 28.2   12.0 ; 28.4  -21.5 ; 28.4   37.5 ;\
         28.6   46.9 ; 29.4  -17.4 ; 30.2   36.2 ; 31.0   75.0 ;\
         31.2    8.1 ; 32.0   54.9 ; 32.0   48.2 ; 32.8   46.9 ;\
         33.4   16.0 ; 33.8   45.6 ; 34.4    1.3 ; 34.8   75.0 ;\
         35.2  -16.0 ; 35.2  -54.9 ; 35.4   69.6 ; 35.6   34.8 ;\
         35.6   32.1 ; 36.2  -37.5 ; 36.2   22.8 ; 38.0   46.9 ;\
         38.0   10.7 ; 39.2    5.4 ; 39.4   -1.3 ; 40.0  -21.5 ;\
         57.6   10.7 ]");

/* deleted data (string too long for the compiler...)
         40.4  -13.3 ; 41.6   30.8 ; 41.6  -10.7 ; 42.4   29.4 ;\
         42.8    0.0 ; 42.8  -10.7 ; 43.0   14.7 ; 44.0   -1.3 ;\
         44.4    0.0 ; 45.0   10.7 ; 46.6   10.7 ; 47.8  -26.8 ;\
         47.8  -14.7 ; 48.8  -13.3 ; 50.6    0.0 ; 52.0   10.7 ;\
         53.2  -14.7 ; 55.0   -2.7 ; 55.0   10.7 ; 55.4   -2.7 ;\
*/

  //x      = data(:,1);
  //t      = data(:,2);
  //scale  = std(t);
  //t      = t/scale;
  //ntp    = size(x,1);
  //X      = (min(x):0.1:max(x))';

  data = ~data;
  arr x         = data[0]; x.reshape(x.N,1);
  arr t         = data[1];
  double scale  = ::sqrt(sumOfSqr(t)/t.N);
  t /= scale;
  arr X;  X.setGrid(1,x.min(),x.max(),100);

  //kernel ridge regression model

  double lambda = 6.;
  double gamma  = 1.;
  arr alpha;
  double b,sigma;
  krr(alpha,b,sigma,
      lambda, gamma, x, t);

  //test
  //mu    = scale*(rbf(lambda, X, x)*alpha + b);
  //sigma = scale*sigma;
  arr mu(X.N),kappa;
  for(uint i=0;i<X.N;i++){
    rbf(kappa,lambda, x, X[i]);
    mu(i)    = scale*(scalarProduct(kappa,alpha) + b);
  }
  sigma *= scale;

  //plot
  t.reshape(t.N,1);
  mu.reshape(mu.N,1);
  arr xsig(X.N,1); xsig=sigma;
  write(x,t*scale,"train");
  write(X,mu,xsig,"test");

  std::cout <<"click on gnuplot..." <<std::endl;
  std::ofstream os("z.plt");
  os
    <<"plot 'train' us 1:2 with points,'test' us 1:2:3 with errorlines\n"
    <<"pause mouse" <<std::endl;
  os.close();
  system("gnuplot z.plt");

  return 0;
}

//===========================================================================
//
// rbf
//

void rbf(arr& K,double lambda, const arr& Xbatch, const arr& x){
  //D = sum(x1.^2,2)*ones(1,size(x2,1))+ones(size(x1,1),1)*sum(x2.^2,2)'-2*x1*x2';
  //K = exp(-D/(lambda^2));
  arr D(Xbatch.N);  for(uint i=0;i<Xbatch.N;i++) D(i) = sqrDistance(Xbatch[i],x);
  K = exp(-D/(lambda*lambda));
}

//===========================================================================
//
// krr
//

void krr(arr& alpha,double& b,double& sigma,
         double lambda, double gamma, const arr& x, const arr& t){
  //K     = rbf(lambda, x, x);
  uint i,n=x.d0;
  arr K(n,n);
  for(i=0;i<n;i++) rbf(K[i](),lambda,x,x[i]);

  //one   = ones(size(t));
  //R     = chol(K + gamma*eye(length(t)));
  //omega = R\(R'\one);
  //nu    = R\(R'\t);
  arr one(n),omega,nu,G; one=1.;
  G = K + gamma*Identity(n);
  lapack_Ainv_b_sym(omega,G,one);
  lapack_Ainv_b_sym(nu,G,t);

  //delta = 1/max(sum(omega),realmin);
  //b     = sum(nu)*delta;
  //alpha = nu - omega*b;
  //sigma = std(t - K*alpha - b);
  double delta = sum(omega); if(delta>1e-10) delta=1./delta; else delta=1e20;
  b = sum(nu)*delta;
  alpha = nu - omega*b;
  sigma = sqrt(sumOfSqr(t - K*alpha - b)/t.N);
}
